#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Connection/Session management module."""

import asyncio
from typing import Awaitable, Dict, Union

from pyee import AsyncIOEventEmitter

from biolib.pyppeteer.pyppeteer.connection import Connection, Message, createProtocolError, rewriteError
from biolib.pyppeteer.pyppeteer.errors import NetworkError
from biolib.pyppeteer.pyppeteer.events import Events


class CDPSession(AsyncIOEventEmitter):
    """Chrome Devtools Protocol Session.

    The :class:`CDPSession` instances are used to talk raw Chrome Devtools
    Protocol:

    * protocol methods can be called with :meth:`send` method.
    * protocol events can be subscribed to with :meth:`on` method.

    Documentation on DevTools Protocol can be found
    `here <https://chromedevtools.github.io/devtools-protocol/>`__.
    """

    def __init__(
        self,
        connection: Union[Connection, 'CDPSession'],
        targetType: str,
        sessionId: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Make new session."""
        super().__init__()
        self._callbacks: Dict[int, asyncio.Future] = {}
        self._connection = connection
        self._targetType = targetType
        self._sessionId = sessionId
        self.loop = loop

    def send(self, method: str, params: dict = None) -> Awaitable:
        """Send message to the connected session.

        :arg str method: Protocol method name.
        :arg dict params: Optional method parameters.
        """
        if not self._connection:
            raise NetworkError(
                f'Protocol Error ({method}): Session closed. Most likely the {self._targetType} has been closed.'
            )
        id_ = self._connection._rawSend({'sessionId': self._sessionId, 'method': method, 'params': params or {},})
        callback = self.loop.create_future()
        callback.method = method
        callback.error = NetworkError()
        self._callbacks[id_] = callback
        return callback

    def _onMessage(self, msg: Message) -> None:
        id_ = msg.get('id')
        callback = self._callbacks.get(id_)
        if id_ and id_ in self._callbacks:
            if msg.get('error'):
                callback.set_exception(
                    createProtocolError(
                        callback.error,  # type: ignore
                        callback.method,  # type: ignore
                        msg,
                    )
                )
            else:
                callback.set_result(msg.get('result'))
            del self._callbacks[id_]
        else:
            if msg.get('id'):
                raise ConnectionError(f'Received unexpected message with no callback: {msg}')
            self.emit(msg.get('method'), msg.get('params'))

    async def detach(self) -> None:
        """Detach session from target.

        Once detached, session won't emit any events and can't be used to send
        messages.
        """
        if not self._connection:
            raise NetworkError('Session already detached. Most likelythe {self._targetType} has been closed')
        await self._connection.send('Target.detachFromTarget', {'sessionId': self._sessionId})

    def _onClosed(self) -> None:
        for cb in self._callbacks.values():
            cb.set_exception(
                rewriteError(
                    cb.error,  # type: ignore
                    f'Protocol error {cb.method}: Target closed.',  # type: ignore
                )
            )
        self._callbacks.clear()
        self._connection = None
        self.emit(Events.CDPSession.Disconnected)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import os
from pathlib import Path
from typing import Any, List, Sequence, Union

from appdirs import AppDirs

__author__ = 'Hiroyuki Takagi, Bernardas Ališauskas, Matt Marcus'
__email__ = 'pyppeteer@protonmail.com'
__version__ = '0.2.2'
__chromium_revision__ = '722234'
__base_puppeteer_version__ = 'v2.1.1'
__pyppeteer_home__ = os.environ.get('PYPPETEER_HOME', AppDirs('pyppeteer').user_data_dir)  # type: str


from biolib.pyppeteer.pyppeteer.browser import Browser  # isort:skip
from biolib.pyppeteer.pyppeteer.browser_fetcher import BrowserFetcher, Platform  # isort:skip
from biolib.pyppeteer.pyppeteer.device_descriptors import devices  # isort:skip
from biolib.pyppeteer.pyppeteer.launcher import ChromeLauncher, FirefoxLauncher, launcher  # isort:skip
from biolib.pyppeteer.pyppeteer.models import BrowserOptions, ChromeArgOptions, Devices, LaunchOptions, Protocol  # isort:skip
from biolib.pyppeteer.pyppeteer.websocket_transport import WebsocketTransport  # isort:skip

# Setup root logger
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


class Pyppeteer:
    def __init__(self, projectRoot: str = None, preferredRevision: str = None):
        self._projectRoot = projectRoot
        self._preferredRevision = preferredRevision
        self._lazyLauncher = None
        self.productName = None

    @property
    def executablePath(self) -> Union[str, Path]:
        return self._launcher.executablePath

    @property
    def product(self) -> str:
        return self._launcher.product

    @property
    def devices(self) -> Devices:
        return devices

    async def launch(self, **kwargs: Union[LaunchOptions, ChromeArgOptions, BrowserOptions]) -> Browser:
        if not self.productName and kwargs:
            self.productName = kwargs.get('product')
        return await self._launcher.launch(**kwargs)

    async def connect(
        self,
        browserWSEndpoint: str = None,
        browserURL: str = None,
        transport: WebsocketTransport = None,
        ignoreHTTPSErrors: bool = False,
        slowMo: float = 0,
        defaultViewport: Protocol.Page.Viewport = None,
    ) -> Browser:
        return await self._launcher.connect(
            browserWSEndpoint=browserWSEndpoint,
            browserURL=browserURL,
            ignoreHTTPSErrors=ignoreHTTPSErrors,
            transport=transport,
            slowMo=slowMo,
            defaultViewport=defaultViewport,
        )

    @property
    def _launcher(self) -> Union[FirefoxLauncher, ChromeLauncher]:
        if not self._lazyLauncher:
            self._lazyLauncher = launcher(
                projectRoot=self._projectRoot, preferredRevision=self._preferredRevision, product=self.productName
            )
        return self._lazyLauncher

    async def defaultArgs(
        self,
        args: Sequence[str] = None,
        devtools: bool = False,
        headless: bool = None,
        userDataDir: str = None,
        **_: Any,
    ) -> List[str]:
        return self._launcher.default_args(args=args, devtools=devtools, headless=headless, userDataDir=userDataDir)

    def createBrowserFetcher(self, platform: Platform = None, host: str = None,) -> BrowserFetcher:
        return BrowserFetcher(projectRoot=self._projectRoot, platform=platform, host=host)


# shortcut methods
async def launch(
    projectRoot: Union[Path, str] = None,
    preferredRevision: str = None,
    **kwargs: Union[LaunchOptions, ChromeArgOptions, BrowserOptions],
) -> Browser:
    return await Pyppeteer(projectRoot, preferredRevision).launch(**kwargs)


async def connect(
    projectRoot: Union[Path, str] = None,
    preferredRevision: str = None,
    browserWSEndpoint: str = None,
    browserURL: str = None,
    transport: WebsocketTransport = None,
    ignoreHTTPSErrors: bool = False,
    slowMo: float = 0,
    defaultViewport: Protocol.Page.Viewport = None,
) -> Browser:
    return await Pyppeteer(projectRoot, preferredRevision).connect(
        browserWSEndpoint=browserWSEndpoint,
        browserURL=browserURL,
        transport=transport,
        ignoreHTTPSErrors=ignoreHTTPSErrors,
        slowMo=slowMo,
        defaultViewport=defaultViewport,
    )


version = __version__
version_info = tuple(int(i) for i in version.split('.'))

__all__ = [
    '__chromium_revision__',
    '__pyppeteer_home__',
    'version',
    'version_info',
    'devices',
    'launch',
    'connect',
    'Pyppeteer',
]



import subprocess

from biolib.biolib_logging import logger, logger_no_user_data
from biolib import utils
from biolib.compute_node.utils import SystemExceptionCodes
from biolib.typing_utils import Callable


class ComputeProcessException(Exception):
    def __init__(self, original_error: Exception, biolib_error_code,
                 # Not using SendSystemExceptionType since importing it leads to many circular import problems
                 # TODO: Fix circular import problems when importing SendSystemExceptionType
                 send_system_exception: Callable[[SystemExceptionCodes], None],
                 may_contain_user_data: bool = True):
        super().__init__()

        if not may_contain_user_data:
            logger_no_user_data.error(str(original_error))
        else:
            logger.error(str(original_error))

        send_system_exception(biolib_error_code)


def log_disk_and_memory_usage_info() -> None:
    if not utils.IS_DEV:
        disk_usage_info = subprocess.run(['df', '-h'], check=False, capture_output=True)
        memory_usage_info = subprocess.run(['free', '-h', '--si'], check=False, capture_output=True)

        logger_no_user_data.debug(disk_usage_info)
        logger_no_user_data.debug(memory_usage_info)



import abc
import json
import os
import shutil
import time
import uuid
from datetime import datetime

import appdirs  # type: ignore

from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.job_worker.cache_types import LfsCacheStateDict, UuidStr, StoragePartition, \
    DockerImageCacheStateDict
from biolib.typing_utils import Dict, List, Optional, Generic, TypeVar


class CacheStateError(BioLibError):
    pass


class DockerCacheStateError(CacheStateError):
    pass


StateType = TypeVar('StateType')


class CacheState(abc.ABC, Generic[StateType]):
    _cache_dir: str = appdirs.user_cache_dir(appname='pybiolib', appauthor='biolib')

    @property
    @abc.abstractmethod
    def _state_path(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _state_lock_path(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_default_state(self) -> StateType:
        raise NotImplementedError

    def __init__(self):
        self._state: Optional[StateType] = None

    def __enter__(self) -> StateType:
        self._acquire_state_lock()
        logger_no_user_data.debug('Acquired state lock')
        logger_no_user_data.debug(f'Does state path {self._state_path} exist? {os.path.exists(self._state_path)}')
        if os.path.exists(self._state_path):
            logger_no_user_data.debug('Opening and reading state path')
            with open(self._state_path, mode='r') as file:
                logger_no_user_data.debug('Opened state path. Loading json')
                self._state = json.loads(file.read())
        else:
            logger_no_user_data.debug('State path does not exist, getting default state')
            self._state = self._get_default_state()
            logger_no_user_data.debug('Opening and reading state path')
            with open(self._state_path, mode='w') as file:
                logger_no_user_data.debug('Opened state path. Dumping json')
                file.write(json.dumps(self._state))

        # Check for type checking
        if self._state is None:
            raise CacheStateError('Internal state is not defined')

        logger_no_user_data.debug(f'Got LFS state {self._state}')

        return self._state

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        with open(self._state_path, mode='w') as file:
            file.write(json.dumps(self._state))

        self._release_state_lock()

    def _acquire_state_lock(self) -> None:
        logger_no_user_data.debug('Acquiring state lock...')
        timeout_seconds = 5.0
        seconds_to_sleep = 0.5
        logger_no_user_data.debug(
            f'Does lock path {self._state_lock_path} exist? {os.path.exists(self._state_lock_path)}'
        )
        while os.path.exists(self._state_lock_path):
            logger_no_user_data.debug('Sleeping until lock path no longer exists')
            time.sleep(seconds_to_sleep)
            timeout_seconds -= seconds_to_sleep
            if timeout_seconds < 0:
                logger_no_user_data.debug('Timed out while waiting for lock file')
                raise CacheStateError('Cache state timed out waiting for lock file')

        logger_no_user_data.debug('Making cache dir')
        os.makedirs(self._cache_dir, exist_ok=True)
        logger_no_user_data.debug('Opening lock file')
        lock_file = open(self._state_lock_path, mode='x')
        logger_no_user_data.debug('Closing log file')
        lock_file.close()
        logger_no_user_data.debug('Closed log file')

    def _release_state_lock(self) -> None:
        if os.path.exists(self._state_lock_path):
            os.remove(self._state_lock_path)
        else:
            raise CacheStateError('Cache state was not locked.')

    @staticmethod
    def get_timestamp_now() -> str:
        return datetime.now().isoformat()


class LfsCacheState(CacheState):
    @property
    def _state_path(self) -> str:
        return f'{super()._cache_dir}/lfs-cache-state.json'

    @property
    def _state_lock_path(self) -> str:
        return f'{self._state_path}.lock'

    def _get_default_state(self) -> LfsCacheStateDict:
        return LfsCacheStateDict(
            large_file_systems={},
            storage_partitions=self._get_storage_partitions_from_env(),
        )

    @staticmethod
    def get_tmp_storage_paths() -> List[str]:
        lfs_tmp_storage_path_env_key = 'BIOLIB_LFS_TMP_STORAGE_PATHS'
        lfs_tmp_storage_paths = os.environ.get(lfs_tmp_storage_path_env_key)
        if lfs_tmp_storage_paths is None:
            raise CacheStateError(f'Environment variable "{lfs_tmp_storage_path_env_key}" not set')

        return lfs_tmp_storage_paths.split(',')

    @staticmethod
    def _get_storage_partitions_from_env() -> Dict[UuidStr, StoragePartition]:
        lfs_storage_path_env_key = 'BIOLIB_LFS_STORAGE_PATHS'
        lfs_storage_paths = os.environ.get(lfs_storage_path_env_key)
        if lfs_storage_paths is None:
            raise CacheStateError(f'Environment variable "{lfs_storage_path_env_key}" not set')

        storage_states: Dict[UuidStr, StoragePartition] = {}
        for lfs_storage_path in lfs_storage_paths.split(','):
            if not os.path.isdir(lfs_storage_path):
                raise CacheStateError(f'LFS storage path {lfs_storage_path} is not a directory')

            uuid_str = str(uuid.uuid4())
            disk_usage = shutil.disk_usage(lfs_storage_path)
            storage_states[uuid_str] = StoragePartition(
                allocated_size_bytes=0,
                path=lfs_storage_path,
                total_size_bytes=disk_usage.free,
                uuid=uuid_str,
            )

        return storage_states


class DockerImageCacheState(CacheState):
    @property
    def _state_path(self) -> str:
        return f'{super()._cache_dir}/docker-cache-state.json'

    @property
    def _state_lock_path(self) -> str:
        return f'{self._state_path}.lock'

    def _get_default_state(self) -> DockerImageCacheStateDict:
        return {}



import os
import shutil
import subprocess
import time
import zipfile

import docker.types  # type: ignore

from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.job_worker.cache_state import LfsCacheState
from biolib.compute_node.job_worker.cache_types import LargeFileSystemCache, StoragePartition
from biolib.typing_utils import TypedDict, Optional, Callable

from biolib.biolib_api_client import LargeFileSystemMapping
from biolib.utils import download_presigned_s3_url


class StatusUpdate(TypedDict):
    progress: int
    log_message: str


class LargeFileSystemAttachResponse(TypedDict):
    aws_ebs_volume_id: str
    device_name: str


class DeviceInfo(TypedDict):
    attached_device_name: str
    aws_ebs_volume_id: str
    nvme_device_name: str


class LargeFileSystemError(BioLibError):
    pass


class LargeFileSystem:

    def __init__(
            self,
            is_job_federated: bool,
            job_id: str,
            lfs_mapping: LargeFileSystemMapping,
            send_status_update: Callable[[StatusUpdate], None],
    ):
        self._is_job_federated = is_job_federated
        self._job_id: str = job_id
        self._lfs_mapping: LargeFileSystemMapping = lfs_mapping
        self._path_on_disk: Optional[str] = None
        self._send_status_update: Callable[[StatusUpdate], None] = send_status_update

    @property
    def _is_initialized(self) -> bool:
        return self._path_on_disk is not None

    @property
    def uuid(self) -> str:
        return self._lfs_mapping['uuid']

    @property
    def docker_mount(self) -> docker.types.Mount:
        if not self._is_initialized:
            raise LargeFileSystemError('LargeFileSystem not initialized')

        return docker.types.Mount(
            read_only=True,
            source=self._path_on_disk,
            target=self._lfs_mapping['to_path'],
            type='bind',
        )

    def initialize(self) -> None:
        logger_no_user_data.debug(f'Initializing LFS {self}...')
        if self._is_initialized:
            logger_no_user_data.debug(f'LFS {self} is already initialized')
            return

        lfs_size_bytes = self._lfs_mapping['size_bytes']
        logger_no_user_data.debug(f'LFS is {lfs_size_bytes} bytes big')

        lfs_is_already_downloading = False

        logger_no_user_data.debug('Accessing LFS Cache...')
        with LfsCacheState() as cache_state:
            logger_no_user_data.debug('Trying to get LFS from cache...')
            lfs_cache = cache_state['large_file_systems'].get(self.uuid)

            if lfs_cache is None:
                logger_no_user_data.debug('LFS was not found in cache...')
                storage_partition_to_use: Optional[StoragePartition] = None
                logger_no_user_data.debug(
                    f'Finding suitable storage partition in partions {cache_state["storage_partitions"].values()}'
                )
                for storage_partition in cache_state['storage_partitions'].values():
                    free_space_bytes = storage_partition['total_size_bytes'] - storage_partition['allocated_size_bytes']
                    logger_no_user_data.debug(
                        f'Checking partition {storage_partition} with free bytes: {free_space_bytes}'
                    )
                    if lfs_size_bytes < free_space_bytes:
                        logger_no_user_data.debug("Using this storage partition")
                        storage_partition_to_use = storage_partition
                        break

                if storage_partition_to_use is None:
                    raise LargeFileSystemError('No storage partition with space available')
                else:
                    storage_partition_to_use['allocated_size_bytes'] += lfs_size_bytes

                logger_no_user_data.debug(
                    'Updating cache with LFS storage partition and size'
                )
                cache_state['large_file_systems'][self.uuid] = LargeFileSystemCache(
                    active_jobs=[self._job_id],
                    last_used_at=LfsCacheState.get_timestamp_now(),
                    size_bytes=lfs_size_bytes,
                    state='downloading',
                    storage_partition_uuid=storage_partition_to_use['uuid'],
                    uuid=self.uuid,
                )

                self._path_on_disk = f"{storage_partition_to_use['path']}/lfs/{self.uuid}/data"

            else:
                logger_no_user_data.debug('LFS found in cache')
                lfs_cache['active_jobs'].append(self._job_id)
                logger_no_user_data.debug('Getting storage partition from Cache')
                storage_partition = cache_state['storage_partitions'][lfs_cache['storage_partition_uuid']]
                self._path_on_disk = f"{storage_partition['path']}/lfs/{self.uuid}/data"

                if lfs_cache['state'] == 'ready':
                    logger_no_user_data.debug('LFS is ready to use')
                    return
                else:
                    logger_no_user_data.debug('LFS is already downloading')
                    lfs_is_already_downloading = True

        logger_no_user_data.debug("Checking if LFS should be downloaded, is downloading or is ready...")
        # TODO: Come up with better status reporting such that the progress values below make sense
        if lfs_is_already_downloading:
            self._send_status_update(StatusUpdate(
                progress=30,
                log_message=f'Waiting for Large File System "{self.uuid}" to be ready...',
            ))
            self._wait_for_lfs_to_be_ready()
            self._send_status_update(StatusUpdate(progress=33, log_message=f'Large File System "{self.uuid}" ready.'))
        else:
            self._send_status_update(StatusUpdate(
                progress=30,
                log_message=f'Downloading Large File System "{self.uuid}"...',
            ))
            self._download_and_unzip()
            self._send_status_update(StatusUpdate(
                progress=33,
                log_message=f'Large File System "{self.uuid}" downloaded.',
            ))
            with LfsCacheState() as cache_state:
                cache_state['large_file_systems'][self.uuid]['state'] = 'ready'

        logger_no_user_data.debug(f"LFS {self} is initialized")

    def detach(self) -> None:
        if not self._is_initialized:
            return

        with LfsCacheState() as cache_state:
            lfs_cache = cache_state['large_file_systems'][self.uuid]
            lfs_cache['last_used_at'] = LfsCacheState.get_timestamp_now()
            lfs_cache['active_jobs'] = [job_id for job_id in lfs_cache['active_jobs'] if job_id != self._job_id]

        self._path_on_disk = None

    def _wait_for_lfs_to_be_ready(self) -> None:
        # Timeout after 15 min
        for _ in range(180):
            time.sleep(5)
            with LfsCacheState() as cache_state:
                if cache_state['large_file_systems'][self.uuid]['state'] == 'ready':
                    return

        raise LargeFileSystemError(f'Waiting for Large File System "{self.uuid}" downloading timed out')

    def _download_and_unzip(self) -> None:
        logger_no_user_data.debug("Starting downloading and unzipping of LFS")
        lfs_size_bytes = self._lfs_mapping['size_bytes']

        tmp_storage_dir: Optional[str] = None
        logger_no_user_data.debug(f"Finding suitable tmp storage path among: {LfsCacheState.get_tmp_storage_paths()}")
        for path in LfsCacheState.get_tmp_storage_paths():
            disk_usage = shutil.disk_usage(path)
            logger_no_user_data.debug(f"Path {path} has disk usage: {disk_usage}")
            if lfs_size_bytes < disk_usage.free:
                logger_no_user_data.debug(f"Choosing path: {path}")
                tmp_storage_dir = path

        if tmp_storage_dir is None:
            raise LargeFileSystemError('No temporary storage available for downloading Large File System')

        tmp_data_zip_path = f'{tmp_storage_dir}/lfs-{self.uuid}-data.zip'

        if self._is_job_federated:
            logger_no_user_data.debug("Job is federated so downloading LFS from presigned link")
            try:
                download_presigned_s3_url(
                    presigned_url=self._lfs_mapping['presigned_download_url'],
                    output_file_path=tmp_data_zip_path,
                )
            except Exception as error:
                raise LargeFileSystemError(
                    f'Failed to download Large File System data.zip got error: {error}'
                ) from error

        else:
            s3_lfs_bucket_name = os.environ.get('BIOLIB_S3_LFS_BUCKET_NAME')
            s3_data_zip_uri = f's3://{s3_lfs_bucket_name}/lfs/versions/{self.uuid}/data.zip'
            logger_no_user_data.debug(f"LFS uri is {s3_data_zip_uri}")

            download_result = subprocess.run(
                ['s5cmd', 'cp', s3_data_zip_uri, tmp_data_zip_path],
                check=False,
                capture_output=True,
            )

            if download_result.returncode != 0:
                raise LargeFileSystemError(
                    f'Failed to download Large File System data.zip: {download_result.stderr.decode()}'
                )

        logger_no_user_data.debug("Extracting LFS Zip...")
        with zipfile.ZipFile(tmp_data_zip_path, 'r') as zip_ref:
            zip_ref.extractall(self._path_on_disk)

        logger_no_user_data.debug("Removing tmp data zip path...")
        os.remove(tmp_data_zip_path)
        logger_no_user_data.debug("Removed tmp data zip path")



import multiprocessing
import os
import time
from urllib.parse import urlparse

import requests
from importlib_metadata import version, PackageNotFoundError

# try fetching version, if it fails (usually when in dev), add default
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data

try:
    BIOLIB_PACKAGE_VERSION = version('pybiolib')
except PackageNotFoundError:
    BIOLIB_PACKAGE_VERSION = '0.0.0'

IS_DEV = os.getenv('BIOLIB_DEV', '').upper() == 'TRUE'

BIOLIB_PACKAGE_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

BIOLIB_CLOUD_ENVIRONMENT = os.getenv('BIOLIB_CLOUD_ENVIRONMENT', '').lower()

BIOLIB_IS_RUNNING_IN_ENCLAVE = BIOLIB_CLOUD_ENVIRONMENT == 'enclave'

IS_RUNNING_IN_CLOUD = BIOLIB_CLOUD_ENVIRONMENT in ('enclave', 'non-enclave')

if BIOLIB_CLOUD_ENVIRONMENT and not IS_RUNNING_IN_CLOUD:
    logger_no_user_data.warning((
        'BIOLIB_CLOUD_ENVIRONMENT defined but does not specify the cloud environment correctly. ',
        'The compute node will not act as a cloud compute node'
    ))

BIOLIB_CLOUD_SKIP_PCR_VERIFICATION = os.getenv('BIOLIB_CLOUD_SKIP_PCR_VERIFICATION', '').upper() == 'TRUE'

BIOLIB_ENABLE_DNS_PROXY = os.getenv('BIOLIB_ENABLE_DNS_PROXY', '').upper() == 'TRUE'

RUN_DEV_JOB_ID = 'run-dev-mocked-job-id'


def get_absolute_container_image_uri(base_url: str, relative_image_uri: str, job_is_federated: bool = False):
    if base_url == 'https://biolib.com' or job_is_federated:
        container_registry_hostname = 'containers.biolib.com'
    elif base_url in ('https://staging-elb.biolib.com', 'https://staging.biolib.com'):
        container_registry_hostname = 'containers.staging.biolib.com'
    else:
        # Expect registry to be accessible on the hostname of base_url if not running on biolib.com
        base_hostname = urlparse(base_url).hostname
        if not base_hostname:
            raise Exception("Could not get hostname from base_url. Tried to get ecr_proxy_uri for image pulling.")
        container_registry_hostname = base_hostname

    return f'{container_registry_hostname}/{relative_image_uri}'


def _download_chunk(input_tuple):
    start, end, presigned_url = input_tuple
    max_upload_retries = 5

    for retry_attempt in range(max_upload_retries):
        logger_no_user_data.debug(f'Attempt number {retry_attempt} for part {start}')
        try:
            response = requests.get(
                url=presigned_url,
                stream=True,
                headers={'range': f'bytes={start}-{end}'},
                timeout=300,  # timeout after 5 min
            )
            if response.ok:
                return_value = response.raw.data
                logger_no_user_data.debug(f'Returning raw data for part {start}')
                return return_value
            else:
                logger_no_user_data.warning(f'Got not ok response when downloading part {start}:{end}. Retrying...')
        except Exception:  # pylint: disable=broad-except
            logger_no_user_data.warning(f'Encountered error when downloading part {start}:{end}. Retrying...')

        time.sleep(2)

    logger_no_user_data.debug(f'Max retries hit, when downloading part {start}:{end}. Exiting...')
    raise BioLibError(f'Max retries hit, when downloading part {start}:{end}. Exiting...')


def download_presigned_s3_url(presigned_url: str, output_file_path: str) -> None:
    chunk_size = 50_000_000

    with requests.get(presigned_url, stream=True, headers={'range': 'bytes=0-1'}) as response:
        file_size = int(response.headers['Content-Range'].split('/')[1])

    chunk_iterator = [(i, i + chunk_size - 1, presigned_url) for i in range(0, file_size, chunk_size)]

    process_pool = multiprocessing.Pool(
        # use 8 cores, unless less is available
        processes=min(16, multiprocessing.cpu_count() - 1),
    )

    bytes_written = 0

    with open(output_file_path, 'ab') as output_file:
        for index, data in enumerate(process_pool.imap(_download_chunk, chunk_iterator)):
            logger_no_user_data.debug(f'Writing part {index} of {file_size} to file...')
            output_file.write(data)

            bytes_written += chunk_size
            approx_progress_percent = min(bytes_written / file_size * 100, 100)
            logger_no_user_data.debug(
                f'Wrote part {index} of {file_size} to file, '
                f'the approximate progress is {round(approx_progress_percent, 2)}%'
            )

    process_pool.close()



"""
This file is modified based on donkersgoed's repository (https://github.com/donkersgoed/nitropepper-enclave-app)
"""

import Crypto
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
from Crypto.Hash import SHA256

# enclave_build will only be present when running on an enclave
import biolib.compute_node.enclave.libnsm as libnsm  # type: ignore # pylint: disable=import-error, no-name-in-module


class NitroSecureModuleUtils:
    """NSM util class."""

    def __init__(self):
        """Construct a new NSMUtil instance."""
        # Initialize the Rust NSM Library
        self._nsm_fd = libnsm.nsm_lib_init()  # pylint:disable=c-extension-no-member
        # Create a new random function `nsm_rand_func`, which
        # utilizes the NSM module.
        self.nsm_rand_func = lambda num_bytes: libnsm.nsm_get_random(  # pylint:disable=c-extension-no-member
            self._nsm_fd, num_bytes
        )

        # Force pycryptodome to use the new rand function.
        # Without this, pycryptodome defaults to /dev/random
        # and /dev/urandom, which are not available in Enclaves.
        self._monkey_patch_crypto(self.nsm_rand_func)

        # Generate a new RSA certificate, which will be used to
        # generate the Attestation document and to decrypt results
        # for KMS Decrypt calls with this document.
        self._rsa_key = RSA.generate(2048)
        self._public_key = self._rsa_key.publickey().export_key('DER')

    def get_attestation_doc(self):
        """Get the attestation document from /dev/nsm."""
        libnsm_att_doc_cose_signed = libnsm.nsm_get_attestation_doc(  # pylint:disable=c-extension-no-member
            self._nsm_fd,
            self._public_key,
            len(self._public_key)
        )
        return libnsm_att_doc_cose_signed

    def decrypt(self, ciphertext):
        """
        Decrypt ciphertext using private key
        """
        cipher = PKCS1_OAEP.new(self._rsa_key, hashAlgo=SHA256)
        plaintext = cipher.decrypt(ciphertext)

        return plaintext

    @classmethod
    def _monkey_patch_crypto(cls, nsm_rand_func):
        """Monkeypatch Crypto to use the NSM rand function."""
        Crypto.Random.get_random_bytes = nsm_rand_func

        def new_random_read(self, n_bytes):  # pylint:disable=unused-argument
            return nsm_rand_func(n_bytes)

        Crypto.Random._UrandomRNG.read = new_random_read  # pylint:disable=protected-access



import io
import tarfile
import subprocess
import time

from docker.models.containers import Container  # type: ignore
from docker.errors import ImageNotFound  # type: ignore
from docker.models.images import Image  # type: ignore
from docker.models.networks import Network  # type: ignore

from biolib import utils
from biolib.compute_node import enclave
from biolib.compute_node.cloud_utils.enclave_parent_types import VsockProxyResponse
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.typing_utils import Optional, List
from biolib.biolib_api_client import RemoteHost
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger


# Prepare for remote hosts with specified port
class RemoteHostExtended(RemoteHost):
    ports: List[int]


class RemoteHostProxy:
    _DOCKER_IMAGE_URI = 'nginx:1.21.1-alpine'
    _TRAFFIC_FORWARDER_PORT_OFFSET = 10000  # Port offset relative to port of a VSOCK proxy

    def __init__(
            self,
            remote_host: RemoteHost,
            public_network: Network,
            internal_network: Optional[Network],
            job_id: Optional[str],
            ports: List[int]
    ):
        # Default to port 443 for now until backend serves remote_hosts with port specified
        self._remote_host: RemoteHostExtended = RemoteHostExtended(
            hostname=remote_host['hostname'],
            ports=ports
        )
        self._public_network: Network = public_network
        self._internal_network: Optional[Network] = internal_network

        if job_id:
            self._name = f"biolib-remote-host-proxy-{job_id}-{self.hostname}"
        else:
            if not utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
                raise Exception('RemoteHostProxy missing argument "job_id"')
            self._name = f"biolib-enclave-remote-host-proxy-{self.hostname}"

        self._container: Optional[Container] = None
        self._enclave_traffic_forwarder_processes: List[subprocess.Popen] = []
        self._enclave_vsock_proxies: List[VsockProxyResponse] = []

    @property
    def hostname(self) -> str:
        return self._remote_host['hostname']

    def get_ip_address_on_network(self, network: Network) -> str:
        if not self._container:
            raise Exception('RemoteHostProxy not yet started')

        container_networks = self._container.attrs['NetworkSettings']['Networks']
        if network.name in container_networks:
            ip_address: str = container_networks[network.name]['IPAddress']
            return ip_address

        raise Exception(f'RemoteHostProxy not connected to network {network.name}')

    def start(self) -> None:
        # TODO: Implement nice error handling in this method

        upstream_server_name = self._remote_host['hostname']
        upstream_server_ports = self._remote_host['ports']

        if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
            # In an enclave the flow is: application -> remote host proxy -> traffic forwarder -> VSOCK proxy -> remote
            upstream_server_name = self._public_network.attrs['IPAM']['Config'][0]['Gateway']
            upstream_server_ports = self._start_vsock_proxy_and_traffic_forwarder_and_return_local_ports()

        docker = BiolibDockerClient.get_docker_client()
        self._container = docker.containers.create(
            detach=True,
            image=self._get_nginx_docker_image(),
            name=self._name,
            network=self._public_network.name,
        )

        self._write_nginx_config_to_container(
            upstream_server_name,
            upstream_server_ports,
        )

        if self._internal_network:
            self._internal_network.connect(self._container.id)

        self._container.start()

        proxy_is_ready = False
        for retry_count in range(1, 5):
            time.sleep(0.5 * retry_count)
            # Use the container logs as a health check.
            # By using logs instead of a http endpoint on the NGINX we avoid publishing a port of container to the host
            if b'start worker process ' in self._container.logs():
                proxy_is_ready = True
                break

        if not proxy_is_ready:
            self.terminate()
            raise Exception('RemoteHostProxy did not start properly')

        self._container.reload()

    def terminate(self):
        # TODO: Implement nice error handling in this method

        if self._container:
            self._container.remove(force=True)

        for process in self._enclave_traffic_forwarder_processes:
            process.terminate()

        for proxy in self._enclave_vsock_proxies:
            CloudUtils.enclave.stop_vsock_proxy(proxy['id'])

    def _get_nginx_docker_image(self) -> Image:
        docker = BiolibDockerClient.get_docker_client()
        try:
            return docker.images.get(self._DOCKER_IMAGE_URI)
        except ImageNotFound:
            logger.debug('Pulling remote host docker image...')
            return docker.images.pull(self._DOCKER_IMAGE_URI)

    def _start_vsock_proxy_and_traffic_forwarder_and_return_local_ports(self) -> List[int]:
        # TODO: Implement nice error handling in this method

        if not utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
            raise Exception('VSOCK proxy and traffic forwarder should only be started in enclave')

        logger.debug(f"Requesting vsock proxies for hostname {self.hostname}")
        traffic_forwarder_local_ports = []
        for port in self._remote_host['ports']:
            enclave_vsock_proxy = CloudUtils.enclave.start_vsock_proxy(
                hostname=self.hostname,
                port=port
            )
            self._enclave_vsock_proxies.append(enclave_vsock_proxy)

            traffic_forwarder_local_port = enclave_vsock_proxy['port'] + self._TRAFFIC_FORWARDER_PORT_OFFSET
            traffic_forwarder_local_ports.append(traffic_forwarder_local_port)
            logger.debug(
                f"Starting traffic forwarder on local port {traffic_forwarder_local_port} to "
                f"VSOCK port {enclave_vsock_proxy['port']} for hostname {self.hostname}"
            )
            self._enclave_traffic_forwarder_processes.append(
                subprocess.Popen([
                    'biolib_traffic_forwarder',
                    str(traffic_forwarder_local_port),
                    str(enclave.PARENT_CID),
                    str(enclave_vsock_proxy['port']),
                ])
            )
        return traffic_forwarder_local_ports

    def _write_nginx_config_to_container(self, upstream_server_name: str, upstream_server_ports: List[int]) -> None:
        if not self._container:
            raise Exception('RemoteHostProxy container not defined when attempting to write NGINX config')

        docker = BiolibDockerClient.get_docker_client()
        nginx_config = '''
events {}
error_log /dev/stdout info;
stream {
    resolver 127.0.0.11 valid=30s;'''
        for idx, upstream_server_port in enumerate(upstream_server_ports):
            nginx_config += f'''
    map "" $upstream_{idx} {{
        default {upstream_server_name}:{upstream_server_port};
    }}

    server {{
        listen          {self._remote_host['ports'][idx]};
        proxy_pass      $upstream_{idx};
    }}

    server {{
        listen          {self._remote_host['ports'][idx]} udp;
        proxy_pass      $upstream_{idx};
    }}'''

        nginx_config += '''
}
'''
        nginx_config_bytes = nginx_config.encode()
        tarfile_in_memory = io.BytesIO()
        with tarfile.open(fileobj=tarfile_in_memory, mode='w:gz') as tar:
            info = tarfile.TarInfo('/nginx.conf')
            info.size = len(nginx_config_bytes)
            tar.addfile(info, io.BytesIO(nginx_config_bytes))

        tarfile_bytes = tarfile_in_memory.getvalue()
        tarfile_in_memory.close()
        docker.api.put_archive(self._container.id, '/etc/nginx', tarfile_bytes)



# pylint: disable=unsubscriptable-object
import os
import subprocess
import time
import signal
from multiprocessing import Process
from datetime import datetime
from socket import gethostbyname, gethostname
import requests

from biolib import utils
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.typing_utils import Optional
from biolib.biolib_api_client import BiolibApiClient
from biolib.compute_node.cloud_utils.enclave_parent_types import VsockProxyResponse
from biolib.compute_node.webserver.webserver_types import WebserverConfig, ComputeNodeInfo, ShutdownTimes


class CloudAutoShutdownTimer(Process):
    def __init__(self, seconds_to_wait: int) -> None:
        super().__init__()
        self._seconds_to_wait = seconds_to_wait

    def run(self) -> None:
        time.sleep(self._seconds_to_wait)
        CloudUtils.deregister_and_shutdown()


class _EnclaveUtils:
    _BASE_URL = 'http://127.0.0.1:5005'

    @staticmethod
    def get_webserver_config():
        response = requests.get(f'{_EnclaveUtils._BASE_URL}/config/', timeout=5)
        return response.json()

    @staticmethod
    def deregister_and_shutdown() -> None:
        requests.post(url=f'{_EnclaveUtils._BASE_URL}/deregister_and_shutdown/', timeout=5)

    @staticmethod
    def start_vsock_proxy(hostname: str, port: int) -> VsockProxyResponse:
        remote_host = {'hostname': hostname, 'port': str(port)}
        response = requests.post(url=f'{_EnclaveUtils._BASE_URL}/vsock_proxy/', json=remote_host, timeout=5)
        vsock_proxy: VsockProxyResponse = response.json()
        return vsock_proxy

    @staticmethod
    def stop_vsock_proxy(vsock_proxy_id: str) -> None:
        requests.delete(url=f'{_EnclaveUtils._BASE_URL}/vsock_proxy/{vsock_proxy_id}/', timeout=5)

    @staticmethod
    def log_message_to_log_file(log_message: str, level: int) -> None:
        requests.post(
            url=f'{_EnclaveUtils._BASE_URL}/log/',
            json={
                'log_message': log_message,
                'level': level,
            },
            timeout=5,
        )


class CloudUtils:
    _webserver_config: Optional[WebserverConfig] = None
    enclave = _EnclaveUtils
    _auto_shutdown_timer_pid_file_name: str = '/tmp/biolib_auto_shutdown_timer.pid'

    @staticmethod
    def initialize() -> None:
        logger_no_user_data.info('Reporting availability...')
        CloudUtils._report_availability()
        CloudUtils.start_auto_shutdown_timer()

    @staticmethod
    def get_webserver_config() -> WebserverConfig:
        if CloudUtils._webserver_config:
            return CloudUtils._webserver_config

        if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
            CloudUtils._webserver_config = CloudUtils.enclave.get_webserver_config()

        else:
            CloudUtils._webserver_config = WebserverConfig(
                compute_node_info=ComputeNodeInfo(
                    auth_token=CloudUtils._get_environment_variable('BIOLIB_COMPUTE_NODE_AUTH_TOKEN'),
                    public_id=CloudUtils._get_environment_variable('BIOLIB_COMPUTE_NODE_PUBLIC_ID'),
                    ip_address=gethostbyname(gethostname())
                ),
                base_url=CloudUtils._get_environment_variable('BIOLIB_BASE_URL'),
                ecr_region_name=CloudUtils._get_environment_variable('BIOLIB_ECR_REGION_NAME'),
                s3_general_storage_bucket_name=CloudUtils._get_environment_variable(
                    'BIOLIB_S3_GENERAL_STORAGE_BUCKET_NAME'
                ),
                s3_lfs_bucket_name=CloudUtils._get_environment_variable('BIOLIB_S3_LFS_BUCKET_NAME'),
                max_docker_image_cache_size_bytes=CloudUtils._get_environment_variable_as_int(
                    'BIOLIB_MAX_DOCKER_IMAGE_CACHE_SIZE_BYTES'
                ),
                is_dev=CloudUtils._get_environment_variable('BIOLIB_DEV').upper() == 'TRUE',
                shutdown_times=ShutdownTimes(
                    job_max_runtime_shutdown_time_in_seconds=CloudUtils._get_environment_variable_as_int(
                        'BIOLIB_CLOUD_JOB_MAX_RUNTIME_IN_SECONDS'
                    ),
                    auto_shutdown_time_in_seconds=CloudUtils._get_environment_variable_as_int(
                        'BIOLIB_CLOUD_AUTO_SHUTDOWN_TIME_IN_SECONDS'
                    ),
                )
            )

        return CloudUtils._webserver_config

    @staticmethod
    def deregister_and_shutdown() -> None:
        logger.debug('Waiting 10 seconds, deregistering and shutting down...')

        # Sleep for 10 seconds to ensure logs are written
        time.sleep(10)

        if utils.BIOLIB_IS_RUNNING_IN_ENCLAVE:
            CloudUtils.enclave.deregister_and_shutdown()
        else:
            config = CloudUtils.get_webserver_config()
            try:
                requests.post(url=f'{config["base_url"]}/api/jobs/deregister/', json={
                    'auth_token': config["compute_node_info"]["auth_token"],
                    'public_id': config["compute_node_info"]["public_id"],
                })
            except Exception as error:  # pylint: disable=broad-except
                logger.error(error)

            logger_no_user_data.debug('Shutting down...')
            try:
                subprocess.run(['sudo', 'shutdown', 'now'], check=True)
            except Exception as error:  # pylint: disable=broad-except
                logger.error(error)

    @staticmethod
    def finish_cloud_job(cloud_job_id: str, system_exception_code: Optional[int]) -> None:
        config = CloudUtils.get_webserver_config()
        try:
            requests.post(
                url=f'{config["base_url"]}/api/jobs/cloud/finish/',
                json={
                    'auth_token': config["compute_node_info"]["auth_token"],
                    'cloud_job_id': cloud_job_id,
                    'system_exception_code': system_exception_code,
                },
                timeout=5,
            )

        except Exception as error:  # pylint: disable=broad-except
            logger_no_user_data.error(f'Could not finish cloud job with id: {cloud_job_id}')
            logger.error(error)

    @staticmethod
    def start_auto_shutdown_timer() -> None:
        if not utils.IS_RUNNING_IN_CLOUD:
            raise BioLibError('Can not start shutdown timer when not running in cloud.')

        if os.path.exists(CloudUtils._auto_shutdown_timer_pid_file_name):
            logger.debug("Re-starting the auto shutdown timer")
            try:
                pid = int(open(CloudUtils._auto_shutdown_timer_pid_file_name, 'r').read())
                os.kill(pid, signal.SIGTERM)

            except Exception as error:  # pylint: disable=broad-except
                logger.warning(error)
                logger.warning('Could not kill old auto shutdown timer')

        config = CloudUtils.get_webserver_config()
        timer_process = CloudAutoShutdownTimer(
            seconds_to_wait=config['shutdown_times']['auto_shutdown_time_in_seconds']
        )

        timer_process.start()

        with open(CloudUtils._auto_shutdown_timer_pid_file_name, 'w') as pid_file:
            pid_file.write(str(timer_process.pid))

    @staticmethod
    def _report_availability() -> None:
        try:
            config = CloudUtils.get_webserver_config()
            compute_node_info = config['compute_node_info']
            api_client = BiolibApiClient.get()
            logger_no_user_data.debug(
                f'Registering with {compute_node_info} to host {api_client.base_url} at {datetime.now()}'
            )

            response: Optional[requests.Response] = None
            max_retries = 5
            for retry_count in range(max_retries):
                try:
                    response = requests.post(f'{api_client.base_url}/api/jobs/report_available/',
                                             json=compute_node_info)
                    break
                except Exception as error:  # pylint: disable=broad-except
                    logger_no_user_data.error(f'Self-registering failed with error: {error}')
                    if retry_count < max_retries - 1:
                        seconds_to_sleep = 1
                        logger_no_user_data.info(f'Retrying self-registering in {seconds_to_sleep} seconds')
                        time.sleep(seconds_to_sleep)

            if not response:
                raise BioLibError('Failed to register. Max retry limit reached')

            if response.status_code != 201:
                raise Exception("Non 201 error code")

        except Exception as exception:  # pylint: disable=broad-except
            logger_no_user_data.error(f'Shutting down as self register failed due to: {exception}')
            if not utils.IS_DEV:
                CloudUtils.deregister_and_shutdown()

    @staticmethod
    def _get_environment_variable(key: str) -> str:
        value = os.environ.get(key)
        # Purposely loose falsy check (instead of `is not None`) as empty string should fail
        if not value:
            raise Exception(f'CloudUtils: Missing environment variable "{key}"')

        return value

    @staticmethod
    def _get_environment_variable_as_int(key: str) -> int:
        return int(CloudUtils._get_environment_variable(key))
